#! /usr/bin/env ruby -Ke
#
# ISPACKޥ˥奢ϤRubyåѡ
# (:ΤȤISPACKޥ˥奢򾯤ʤȤޤưʤ)
#
# history:
# 05/05/20 ispack_wrap.rb 
# 05/06/20   ver.11 Ϥ¿ʤ1롢ΰĹɬפʬְʾ
# 05/06/24   ver.12 ̥åѡ
# 05/07/20   ver.13 Ĺåshapetotalѹ1(ver11)ϼ
# 05/08/05   ver.14 ispack-0.71򻲾Ȥ褦ѹ
# 05/10/03   ver.15 P2PACK,SPPACK,SOPACKθб
# 05/10/12   ver.16 SMPACK,LTPACKΥХ
#                   CåѡκˡѹC2SWCK,SPSWNX,SPSWHXΥХfix
# 06/01/25 ispack_makewrapper.rb ̾FTPACKΥХ
#-----------------------------------------------------------------------------#

libfilename = "ispack_raw" # .so ˻Ȥ̾
rbfilename = "ispack_.rb" # Ruby̥åѡե̾
modulename = "ISPACK" # RubyץǻȤ⥸塼̾
srcpath = "/home/rmizuta/install/ispack-0.71" # ISPACKΥǥ쥯ȥ

cwrapper = true # Cåѡ񤭽Ф
rubywrapper = false # Ruby̥åѡǤ񤭽Ф(ޤʲϤΤ)

filenames = [
#  srcpath+"/appack/doc/appack.tex", 
#  srcpath+"/bspack/doc/bspack.tex", # ʸ̤б
#  srcpath+"/fepack/doc/fepack.tex", # եϤ̤б(̵)
#  srcpath+"/fhpack/doc/fhpack.tex", # Ʊ
#  srcpath+"/flpack/doc/flpack.tex", # ʸ̤б
  srcpath+"/stpack/doc/stpack.tex",
#  srcpath+"/tdpack/doc/tdpack.tex", # ֥롼ˤǤ롩
  srcpath+"/smpack/doc/smpack.tex",
#  srcpath+"/snpack/doc/snpack.tex", # 
  "./snpack.tex", 
#  srcpath+"/sopack/doc/sopack.tex", # 
  "./sopack.tex", 
#  srcpath+"/c2pack/doc/c2pack.tex", # ̤ĤƤʤΤ롢(6/17)
  "./c2pack.tex",
  srcpath+"/ftpack/doc/ftpack.tex",  # ޥ˥奢뤬ü
  srcpath+"/n2pack/doc/n2pack.tex",
#  srcpath+"/sppack/doc/sppack.tex", # 
  "./sppack.tex",
#  srcpath+"/p2pack/doc/p2pack.tex", # 
  "./p2pack.tex",
  srcpath+"/p3pack/doc/p3pack.tex",
  "./ndpack.tex"                   # ʤΤǺä
]

=begin

filenames = [ # for debug
#  "ispack/stpack.tex",
#  "ispack/ndpack.tex",
  "./snpack.tex", 
]

=end

routinelist = Array.new
varlist = Hash.new
vars1 = Hash.new
vars = Hash.new

filenames.each{ |filename|

file = File.open(filename,"r")

loop do
  line = file.gets
  break if line == nil

#  if line =~ /^\\subsection\{(.*)\}$/
#  if line =~ /^\\subsection\{([^\/]+)\/*.*\}$/ # for STPACK
#  if line =~ /^\\subsection\{(.{6,6})\}$/ || line =~ /^\\subsection\{([^\/]+)\/.*\}$/ # for STPACK,SNPACK
  if line =~ /^\\subsection\{(.{6,6})\}$/
    @subroutine = $1
    p @subroutine
    routinelist.push(@subroutine)

    @subroutine1 = nil
    @subroutine2 = nil

  elsif line =~ /^\\subsection\{(.{6,6})\/(.{6,6})\}$/ # for ST,LTPACK
    @subroutine = $1
    p @subroutine
    routinelist.push(@subroutine)

    @subroutine1 = $2
    p @subroutine1
    routinelist.push(@subroutine1)

    @subroutine2 = nil

  elsif line =~ /^\\subsection\{(.{6,6})\/(.{6,6})\/(.{6,6})\}$/ # for ST,LT,FT

    @subroutine = $1
    p @subroutine
    routinelist.push(@subroutine)

    @subroutine1 = $2
    p @subroutine1
    routinelist.push(@subroutine1)

    @subroutine2 = $3
    p @subroutine2
    routinelist.push(@subroutine2)

  end
#  p @subroutine

#  if line =~ /^\{\\tt #{@subroutine}\((.*)\)\}$/
#  if line =~ /^.*\{\\tt #{@subroutine}\((.*)\)\}(\\\\)*$/ # for ST,LT,FTPACK
  if line =~ /^.*\{\\tt #{@subroutine}\(([^\)]*)(\))*(\})*(\\\\)*$/ # for ST,LT,FT,C2,SP
    list = $1
    tail1 = $2
    tail2 = $3
    tail3 = $4

    if ( !tail1 ) # for C2PACK
      line = line.chomp + file.gets # 
#      p line
      if line =~ /^.*\{\\tt #{@subroutine}\((.*)\)\}(\\\\)*$/ # for ST,LT,FT,C2,SP
        list = $1
#        @varlist = list.split(",")
        @varlist = list.split(/[,| ]/)
#        p @varlist
      else
        raise "invalid lines? "+line
      end
    else
      @varlist = list.split(",")
    end

#    @varlist = $1.split(",")
#    p @varlist
    varlist[@subroutine] = @varlist
    vars1[@subroutine] = Array.new
    vars[@subroutine] = Hash.new

#    if sub1 # ιԤɤ
    if tail3 # ιԤɤ
#      p @subroutine1
      line = file.gets # ιԤɤ
      if line =~ /^.*\{\\tt #{@subroutine1}\((.*)\)\}(\\\\)*$/
        sub2 = $2
        @varlist = $1.split(",")
        varlist[@subroutine1] = @varlist
#        p varlist[@subroutine1]
        vars1[@subroutine1] = Array.new
        vars[@subroutine1] = Hash.new

        if sub2 # ιԤɤ
#          p @subroutine2
          line = file.gets # ιԤɤ
          if line =~ /^.*\{\\tt #{@subroutine2}\((.*)\)\}(\\\\)*$/
            @varlist = $1.split(",")
            varlist[@subroutine2] = @varlist
#            p varlist[@subroutine2]
            vars1[@subroutine2] = Array.new
            vars[@subroutine2] = Hash.new
          end
        end
      end
    end
  end

#  if line =~ /^\{\\tt (.*)\} *& *\{\\tt (.*)\} *& *$/
#      line = line.chomp + file.gets
#      p line
#  end

  if line =~ /^\{\\tt (.*)\} +& +\{\\tt (.*)\}/
    unless line =~ /^\{\\tt (.*)\} +& +\{\\tt (.*)\}.*\\\\.*$/ # ³
#      p line
      line = line.chomp + file.gets
#      p line
    end
  end


#  if line =~ /^\{\\tt (.*)\} +& +\{\\tt (.*)\} +& +([к])/
#  if line =~ /^ *\{\\tt +([^ ]+) *\} *& *\{\\tt (.*)\} *& *([к])/ # for FTPACK
  if line =~ /^ *\{\\tt +([^ ]+) *\} *& *\{\\tt (.*)\} *& *\t*([к])/ # for FT,SMPACK
    @name = $1
    type = $2
    case($3)
    when("")
      inout = "in"
    when("")
      inout = "out"
    when("")
      inout = "work"
#      when -------------------ξѿ
    else
      raise "unsupported inout type"
    end

    # ̤ĤƤʤȤ˳̤­
    count = type.count("(") - type.count(")")
    if count > 0
#      p type
      type = type + ")"*count
    elsif count < 0
#      p type
      type = "("*count + type
    end

    if line =~ /^ *\{\\tt .*\} *& *\{\\tt .*\} *&.*礭(.+)/ # for FTPACK
      type1 = $1
      if type == "(D)" || type == "(I)"
#        p type
        type = type.gsub(')','(')
        type += type1.gsub(/^\\tt/,"").gsub("$\\times$","*").gsub("\\tt","*").gsub("**","*")
        type += "))"
#        p type
      else
        raise "unsupported variable type"
      end
    end

#    if @varlist.index(@name) == nil
    if varlist[@subroutine].index(@name) == nil

#      p varlist[@subroutine]
#      p @name

#      raise "#{@name} is not listed in the arguments"
#      print "#{@name} is not listed in the arguments of #{@subroutine}\n" ## for debug
    else
      var = {"name"=>@name.downcase, "type"=>type, "inout"=>inout}
      vars1[@subroutine].push(var)
      vars[@subroutine][@name.downcase] = {"type"=>type, "inout"=>inout}
    end

    if @subroutine1
      if varlist[@subroutine1].index(@name) == nil
#        print "#{@name} is not listed in the arguments of #{@subroutine1}\n" ## for debug
      else
        var = {"name"=>@name.downcase, "type"=>type, "inout"=>inout}
        vars1[@subroutine1].push(var)
        vars[@subroutine1][@name.downcase] = {"type"=>type, "inout"=>inout}
      end
    end

    if @subroutine2
      if varlist[@subroutine2].index(@name) == nil
#        print "#{@name} is not listed in the arguments of #{@subroutine2}\n" ## for debug
      else
        var = {"name"=>@name.downcase, "type"=>type, "inout"=>inout}
        vars1[@subroutine2].push(var)
        vars[@subroutine2][@name.downcase] = {"type"=>type, "inout"=>inout}
      end
    end

#    var = {"name"=>name.downcase, "type"=>type, "inout"=>inout}
#    vars1[@subroutine].push(var)

  end

  if line =~ /^ *& +& +([к])/ # ξѿ
#    p line
    inout1 = $1
    if vars1[@subroutine][-1]["name"] == @name.downcase
      inout = vars1[@subroutine][-1]["inout"]
#      if (inout == "in" && $1 =~ //) || (inout == "out" && $1 == //)
      if (inout == "in" && inout1 =~ //) || (inout == "out" && inout1 == //)
        vars1[@subroutine][-1]["inout"] = "inout"
      else
        raise "unsupported inout type"
      end
#    else
#      p vars1[@subroutine][-1]["name"]
#      p @name.downcase
    end
    if @subroutine1 && vars1[@subroutine1][-1]["name"] == @name.downcase
      inout = vars1[@subroutine1][-1]["inout"]
#      if (inout == "in" && $1 =~ //) || (inout == "out" && $1 == //)
      if (inout == "in" && inout1 =~ //) || (inout == "out" && inout1 == //)
        vars1[@subroutine1][-1]["inout"] = "inout"
      else
        raise "unsupported inout type"
      end
    end
    if @subroutine2 && vars1[@subroutine2][-1]["name"] == @name.downcase
      inout = vars1[@subroutine2][-1]["inout"]
#      if (inout == "in" && $1 =~ //) || (inout == "out" && $1 == //)
      if (inout == "in" && inout1 =~ //) || (inout == "out" && inout1 == //)
        vars1[@subroutine2][-1]["inout"] = "inout"
      else
        p inout
        raise "unsupported inout type"
      end
    end

    if vars[@subroutine][@name.downcase]
      inout = vars[@subroutine][@name.downcase]["inout"]
      if (inout == "in" && inout1 =~ //) || (inout == "out" && inout1 == //)
        vars[@subroutine][@name.downcase]["inout"] = "inout"
      else
        raise "unsupported inout type"
      end
    end

    if @subroutine1
      inout = vars[@subroutine1][@name.downcase]["inout"]
      if (inout == "in" && inout1 =~ //) || (inout == "out" && inout1 == //)
        vars[@subroutine1][@name.downcase]["inout"] = "inout"
      else
        raise "unsupported inout type"
      end
    end

    if @subroutine2
      inout = vars[@subroutine2][@name.downcase]["inout"]
      if (inout == "in" && inout1 =~ //) || (inout == "out" && inout1 == //)
        vars[@subroutine2][@name.downcase]["inout"] = "inout"
      else
        raise "unsupported inout type"
      end
    end

  end

end

#p routinelist
#p varlist
#p vars1
#routinelist.each{ |subroutine| 
#  p subroutine
#  p varlist[subroutine].length
#  p vars1[subroutine].length
#}

file.close

} # end of filenames


# 㳰

=begin
vars1["SPSWNW"] = vars1["SPSWNV"] if vars1["SPSWNV"]
vars1["SPSWNX"] = vars1["SPSWNV"] if vars1["SPSWNV"]
vars1["SPSWHW"] = vars1["SPSWHV"] if vars1["SPSWHV"]
vars1["SPSWHX"] = vars1["SPSWHV"] if vars1["SPSWHV"]
=end

vars["SPSWNW"] = vars["SPSWNV"] if vars["SPSWNV"]
vars["SPSWNX"] = vars["SPSWNV"] if vars["SPSWNV"]
vars["SPSWHW"] = vars["SPSWHV"] if vars["SPSWHV"]
vars["SPSWHX"] = vars["SPSWHV"] if vars["SPSWHV"]

#p vars1["LTS2GZ"]
["LTS2GZ","LTS2GW","LTS2VZ","LTS2VW","LTSRVZ","LTSRVW"].each{ |subroutine|
#  vars1[subroutine].push({"name"=>"im", "type"=>"(I)", "inout"=>"in"}) if vars1[subroutine]
#  vars[subroutine]["im"] = {"type"=>"(I)", "inout"=>"in"} if vars[subroutine]
#  varlist[subroutine].push("im") if vars[subroutine]
}

=begin
vars1["C2S2GA"] = vars1["C2S2GA"][0..3] + 
  [{"name"=>"s", "type"=>"(D((2*KM+1)*LM))", "inout"=>"in"}] + 
  vars1["C2S2GA"][4..-1] if vars1["C2S2GA"]
vars1["C2G2SA"] = vars1["C2G2SA"][0..4] + 
  [{"name"=>"s", "type"=>"(D((2*KM+1)*(LM+1)))", "inout"=>"out"}] + 
  vars1["C2G2SA"][5..-1] if vars1["C2G2SA"]
=end

vars["C2S2GA"]["s"] = {"type"=>"(D((2*KM+1)*LM))", "inout"=>"in"} if vars1["C2S2GA"]
vars["C2G2SA"]["s"] = {"type"=>"(D((2*KM+1)*(LM+1)))", "inout"=>"out"} if vars1["C2G2SA"]

["FTTZLI","FTTZUI","FTTRUI","FTTCTI","FTTSTI","FTTCMI","FTTSMI"].each{ |name|
  vars[name]["it"]["inout"] = "out" if vars1[name]
  vars[name]["t"]["inout"] = "out" if vars1[name]
}

["FTTZLM","FTTZUF","FTTZUB"].each{ |name|
  vars[name]["x"]["type"] = "(D(M*N*2))" if vars1[name]
  vars[name]["y"]["type"] = "(D(M*N*2))" if vars1[name]
}

["FTTRUF","FTTRUB","FTTCTF","FTTCTB","FTTSTF","FTTSTB","FTTCMF","FTTCMB","FTTSMF","FTTSMB"].each{ |name|
  vars[name]["x"]["type"] = "(D(M*N))" if vars1[name]
  vars[name]["y"]["type"] = "(D(M*N))" if vars1[name]
}


# 㳰ޤ


#-----------------------------------------------------------------------------#
if(cwrapper) # Cåѡ񤭽Ф

#file = File.open("temp.c","w")
file = File.open(libfilename+".c","w")

file << '#include "ruby.h"'+"\n"
file << '#include "narray.h"'+"\n"
#file << 'VALUE na_flatten_bang(VALUE self);'+"\n"
file << "\n"

definelist = ""

routinelist.each{ |subroutine|
  if varlist[subroutine] == nil
    raise "varlist for #{subroutine} not exist"
  end
#  p varlist[subroutine].length
#  p vars1[subroutine].length
#  p varlist[subroutine]
#  p vars1[subroutine].each{|v| p v["name"]}

#  varlist[subroutine].each{ |varname|
#    p varname
#    p vars1[varname]
#    if vars1[varname]
#  }

=begin

#  inlist = "self"
  invars = Array.new
  outvars = Array.new
  declare = ""
  declare_c2r = nil
  r2c = "\n"
  c2r = "\n"
  args = ""


#p vars1[subroutine]

  vars1[subroutine].each{|v|
#  p v["inout"]
#  if v["inout"]=="in" || v["inout"]=="inout"
    vname = v["name"]
    if v["inout"]=="in" || v["inout"]=="inout" || v["inout"]=="work"
#      inlist = inlist+","+vname
      invars.push(vname)
#    p v["type"]
      outvars.push(vname) if v["inout"]=="inout"
      case(v["type"])
      when("(I)")
        declare += "    int #{vname}1;\n"
        r2c += "    #{vname}1 = NUM2INT(#{vname});\n"
        args += "&#{vname}1,"
      when("(D)")
        declare += "    double #{vname}1;\n"
        r2c += "    #{vname}1 = NUM2DBL(#{vname});\n"
        args += "&#{vname}1,"
      when(/\(I\(/)
#        arylength = v["type"].downcase.sub(/\(i/,"").sub(/\)$/,"").gsub(/([a-z][a-z|0-9]*)/,'\1'+"1")
        arylength = v["type"].downcase.sub(/\(i/,"").sub(/\)$/,"").gsub(/([a-z][a-z|0-9]*)/,'\1'+"1").gsub(/,/,"*")
        declare += "    struct NARRAY *#{vname}1;\n"
        r2c += "    #{vname} = na_cast_object(#{vname}, NA_LINT);\n"
        r2c += "    GetNArray(#{vname},#{vname}1);\n"
#        r2c += "    if ( #{vname}1->rank > 1 ) { \n"
#        r2c += "      #{vname}1->rank = 1;\n"
#        r2c += "      #{vname}1->shape[0] = #{vname}1->total;\n"
#        r2c += "    }\n"
        if v["inout"]=="work"
          r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
        else
          r2c += "    if ( #{vname}1->total != #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d != #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
        end
        args += v["name"]+"1->ptr,"
      when(/\(D\(/)
        arylength = v["type"].downcase.sub(/\(d/,"").sub(/\)$/,"").gsub(/([a-z][a-z|0-9]*)/,'\1'+"1").gsub(/,/,"*")
        declare += "    struct NARRAY *#{vname}1;\n"
        r2c += "    #{vname} = na_cast_object(#{vname}, NA_DFLOAT);\n"
        r2c += "    GetNArray(#{vname},#{vname}1);\n"
#        r2c += "    if ( #{vname}1->rank > 1 ) { \n"
#        r2c += "      #{vname}1->rank = 1;\n"
#        r2c += "      #{vname}1->shape[0] = #{vname}1->total;\n"
#        r2c += "    }\n"
        args += v["name"]+"1->ptr,"
#        args += "(double*)"+v["name"]+"1->ptr,"


        # 㳰

#        if( subroutine == "STOGRD" && vname == "q" )
        if( subroutine =~ /TOGRD/ && vname == "q" )
          arylength = "jm1"  # ˤ
          r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
        elsif( subroutine =~ /^LT/ && vname =~ /[uvgp]/ )
          arylength = "jm1"  # ˤ
          r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
        elsif( (subroutine == "NDTV2A" || subroutine == "NDTA2V" ) && vname == "vrt" )
          arylength = "2"  # ˤ
          r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
        elsif( (subroutine == "C2S2GA" ) && vname == "s" )
          arylength = "(2*km1+1)*lm1"
          r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
        elsif( arylength =~ /礭ηͻ/ )
          if( subroutine =~ /^SN/ )
#            arylength = "km1*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "km1*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
            arylength = "km1*(jd1*im1)" # 3
          elsif( subroutine =~ /^SO/ )
#            arylength = "2*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "2*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
            arylength = "2*(jd1*im1)" # 3
          elsif( subroutine == "SPNJCB" )
#            arylength = "3*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "3*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
            arylength = "3*(jd1*im1)" # 3
          elsif( subroutine == "SPSWNL" )
#            arylength = "5*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "5*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
            arylength = "5*(jd1*im1)" # 3
          elsif( subroutine == "SPSWCK" )
#            arylength = "4*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "4*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
            arylength = "4*(jd1*im1)" # 3
          elsif( subroutine == "SPSWNV" && vname == "ww" )
#            arylength = "((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "(jd1*((mm1+1)/2+mm1+1)*2)" # 2
            arylength = "(jd1*im1)" # 3
          elsif( subroutine == "SPSWNV" && vname == "ws" )
#            arylength = "8*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "8*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
            arylength = "8*(jd1*im1)" # 3
          elsif( subroutine =~ /^SPSWC/ && vname == "ww" )
#            arylength = "((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "(jd1*((mm1+1)/2+mm1+1)*2)" # 2
            arylength = "(jd1*im1)" # 3
          elsif( subroutine =~ /^SPSWC/ && vname == "ws" )
#            arylength = "8*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "8*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
            arylength = "4*(jd1*im1)" # 3
          elsif( subroutine == "SPSWHV" && vname == "ww" )
#            arylength = "((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "(jd1*((mm1+1)/2+mm1+1)*2)" # 2
            arylength = "(jd1*im1)" # 3
          elsif( subroutine == "SPSWHV" && vname == "ws" )
#            arylength = "8*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "8*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
            arylength = "11*(jd1*im1)" # 3
          elsif( subroutine == "SPSWNW" )
            arylength = "8*(jd1*im1)" # 3
          elsif( subroutine == "SPSWNX" )
            arylength = "7*(jd1*im1)" # 3
          elsif( subroutine == "SPSWHW" )
            arylength = "11*(jd1*im1)" # 3
          elsif( subroutine == "SPSWHX" )
            arylength = "10*(jd1*im1)" # 3
          end
          r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
#          r2c += "    if ( #{vname}1->total > #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} (%d > #{arylength}) \", #{vname}1->total );\n"


        # 㳰ޤ



        elsif v["inout"]=="work"
          r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
        else
          r2c += "    if ( #{vname}1->total != #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d != #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
        end

        r2c.gsub!("lmnm1","(2*nm1+1)*(2*mm1+1)*(2*lm1+1)") # for P3PACK


      else
#      raise "#{v["type"]} not supported"
      end
    elsif v["inout"]=="out"
#      if outlist == ""
#        outlist = v["name"]
#      else
#        outlist = outlist+","+v["name"]
#      end
      outvars.push(v["name"])

      case(v["type"])
      when("(I)")
        declare += "    int #{vname}1;\n"
        declare += "    VALUE #{vname};\n"
        args += "&#{vname}1,"
        c2r += "    #{vname} = INT2NUM(#{vname}1);\n"
      when("(D)")
        declare += "    double #{vname}1;\n"
        declare += "    VALUE #{vname};\n"
        args += "&#{vname}1,"
        c2r += "    #{vname} = rb_float_new(#{vname}1);\n"
      when(/\([DI]\(/)
#      p v["type"]
        if v["type"] =~ /\(D\(/
          arytype = "NA_DFLOAT"
          arylength = v["type"].downcase.sub(/\(d/,"").sub(/\)$/,"").gsub(/([a-z][a-z|0-9]*)/,'\1'+"1").gsub(/,/,"*")
          args += v["name"]+"1->ptr,"
#          args += "(double*)"+v["name"]+"1->ptr,"
        elsif v["type"] =~ /\(I\(/
          arytype = "NA_LINT"
          arylength = v["type"].downcase.sub(/\(i/,"").sub(/\)$/,"").gsub(/([a-z][a-z|0-9]*)/,'\1'+"1").gsub(/,/,"*")
          args += v["name"]+"1->ptr,"
        else
          raise "error in determing arytype"
        end
#      p arylength
        declare += "    struct NARRAY *#{vname}1;\n"
        declare += "    VALUE #{vname};\n"
        declare_c2r = "    int rank=1,*shape;\n    shape = ALLOCA_N(int, rank);\n"
        r2c += "    shape[0] = "+arylength+";\n"
        r2c += "    #{vname} = na_make_object( "+arytype+", rank, shape, cNArray );\n"
        r2c += "    GetNArray(#{vname},#{vname}1);\n"
#        args += v["name"]+"1->ptr,"

        r2c.gsub!("lmnm1","(2*nm1+1)*(2*mm1+1)*(2*lm1+1)") # for P3PACK

      else
#      raise "#{v["type"]} not supported"
      end

    end
  } # end of vars1[subroutine].each

=end

  invars = Array.new
  outvars = Array.new
  declare = ""
  declare_c2r = nil
  r2c = "\n"
  c2r = "\n"
  args = ""

  varlist[subroutine].each{|v|
    vname = v.downcase
    if vname =~ /hspace/
      # do nothing
    elsif !vars[subroutine][vname]
      p vars[subroutine]
      p vars1[subroutine]
      raise "vars[#{subroutine}][#{vname}] not exist"
    else
#      p vname
      inout = vars[subroutine][vname]["inout"]
      type = vars[subroutine][vname]["type"]
      if inout =="in" || inout =="inout" || inout =="work"
        invars.push(vname)
        outvars.push(vname) if inout =="inout"
        case(type)
        when("(I)")
          declare += "    int #{vname}1;\n"
          r2c += "    #{vname}1 = NUM2INT(#{vname});\n"
          args += "&#{vname}1,"
        when("(D)")
          declare += "    double #{vname}1;\n"
          r2c += "    #{vname}1 = NUM2DBL(#{vname});\n"
          args += "&#{vname}1,"
        when(/\(I\(/)
          arylength = type.downcase.sub(/\(i/,"").sub(/\)$/,"").gsub(/([a-z][a-z|0-9]*)/,'\1'+"1").gsub(/,/,"*")
          declare += "    struct NARRAY *#{vname}1;\n"
          r2c += "    #{vname} = na_cast_object(#{vname}, NA_LINT);\n"
          r2c += "    GetNArray(#{vname},#{vname}1);\n"
          if inout =="work"
            r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
          else
            r2c += "    if ( #{vname}1->total != #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d != #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
          end
          args += vname+"1->ptr,"
        when(/\(D\(/)
          arylength = type.downcase.sub(/\(d/,"").sub(/\)$/,"").gsub(/([a-z][a-z|0-9]*)/,'\1'+"1").gsub(/,/,"*")
          declare += "    struct NARRAY *#{vname}1;\n"
          r2c += "    #{vname} = na_cast_object(#{vname}, NA_DFLOAT);\n"
          r2c += "    GetNArray(#{vname},#{vname}1);\n"
          args += vname+"1->ptr,"

          # 㳰

          if( subroutine =~ /TOGRD/ && vname == "q" )
            arylength = "jm1"  # ˤ
            r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
          elsif( subroutine =~ /^LT/ && vname =~ /[uvgp]/ )
            arylength = "jm1"  # ˤ
            r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
          elsif( (subroutine == "NDTV2A" || subroutine == "NDTA2V" ) && vname == "vrt" )
            arylength = "2"  # ˤ
            r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
          elsif( (subroutine == "C2S2GA" ) && vname == "s" )
            arylength = "(2*km1+1)*lm1"
            r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
          elsif( arylength =~ /礭ηͻ/ )
            if( subroutine =~ /^SN/ )
#            arylength = "km1*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "km1*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
              arylength = "km1*(jd1*im1)" # 3
            elsif( subroutine =~ /^SO/ )
#            arylength = "2*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "2*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
              arylength = "2*(jd1*im1)" # 3
            elsif( subroutine == "SPNJCB" )
#            arylength = "3*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "3*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
              arylength = "3*(jd1*im1)" # 3
            elsif( subroutine == "SPSWNL" )
#            arylength = "5*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "5*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
              arylength = "5*(jd1*im1)" # 3
            elsif( subroutine == "SPSWCK" )
#            arylength = "4*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "4*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
              arylength = "4*(jd1*im1)" # 3
            elsif( subroutine == "SPSWNV" && vname == "ww" )
#            arylength = "((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "(jd1*((mm1+1)/2+mm1+1)*2)" # 2
              arylength = "(jd1*im1)" # 3
            elsif( subroutine == "SPSWNV" && vname == "ws" )
#            arylength = "8*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "8*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
              arylength = "8*(jd1*im1)" # 3
            elsif( subroutine =~ /^SPSWC/ && vname == "ww" )
#            arylength = "((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "(jd1*((mm1+1)/2+mm1+1)*2)" # 2
              arylength = "(jd1*im1)" # 3
            elsif( subroutine =~ /^SPSWC/ && vname == "ws" )
#            arylength = "8*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "8*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
              arylength = "4*(jd1*im1)" # 3
            elsif( subroutine == "SPSWHV" && vname == "ww" )
#            arylength = "((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "(jd1*((mm1+1)/2+mm1+1)*2)" # 2
              arylength = "(jd1*im1)" # 3
            elsif( subroutine == "SPSWHV" && vname == "ws" )
#            arylength = "8*((mm1+1)/2*2+3)*(mm1/2+2)*2" # 1
#            arylength = "8*(jd1*((mm1+1)/2+mm1+1)*2)" # 2
              arylength = "11*(jd1*im1)" # 3
            elsif( subroutine == "SPSWNW" )
              arylength = "8*(jd1*im1)" # 3
            elsif( subroutine == "SPSWNX" )
              arylength = "7*(jd1*im1)" # 3
            elsif( subroutine == "SPSWHW" )
              arylength = "11*(jd1*im1)" # 3
            elsif( subroutine == "SPSWHX" )
              arylength = "10*(jd1*im1)" # 3
            end
            r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"

          # 㳰ޤ

          elsif inout =="work"
            r2c += "    if ( #{vname}1->total < #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d < #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
          else
            r2c += "    if ( #{vname}1->total != #{arylength} ) rb_raise(rb_eArgError, \"invalid ary size: #{vname} : %d != #{arylength} = %d \", #{vname}1->total, #{arylength} );\n"
          end

          r2c.gsub!("lmnm1","(2*nm1+1)*(2*mm1+1)*(2*lm1+1)") # for P3PACK
        end
      elsif inout =="out"
        outvars.push(vname)

        case(type)
        when("(I)")
          declare += "    int #{vname}1;\n"
          declare += "    VALUE #{vname};\n"
          args += "&#{vname}1,"
          c2r += "    #{vname} = INT2NUM(#{vname}1);\n"
        when("(D)")
          declare += "    double #{vname}1;\n"
          declare += "    VALUE #{vname};\n"
          args += "&#{vname}1,"
          c2r += "    #{vname} = rb_float_new(#{vname}1);\n"
        when(/\([DI]\(/)
          if type =~ /\(D\(/
            arytype = "NA_DFLOAT"
            arylength = type.downcase.sub(/\(d/,"").sub(/\)$/,"").gsub(/([a-z][a-z|0-9]*)/,'\1'+"1").gsub(/,/,"*")
            args += vname+"1->ptr,"
          elsif type =~ /\(I\(/
            arytype = "NA_LINT"
            arylength = type.downcase.sub(/\(i/,"").sub(/\)$/,"").gsub(/([a-z][a-z|0-9]*)/,'\1'+"1").gsub(/,/,"*")
            args += vname+"1->ptr,"
          else
            raise "error in determing arytype"
          end
          declare += "    struct NARRAY *#{vname}1;\n"
          declare += "    VALUE #{vname};\n"
          declare_c2r = "    int rank=1,*shape;\n    shape = ALLOCA_N(int, rank);\n"
          r2c += "    shape[0] = "+arylength+";\n"
          r2c += "    #{vname} = na_make_object( "+arytype+", rank, shape, cNArray );\n"
          r2c += "    GetNArray(#{vname},#{vname}1);\n"

          r2c.gsub!("lmnm1","(2*nm1+1)*(2*mm1+1)*(2*lm1+1)") # for P3PACK

        else

        end

      end

    end
  }

  case(subroutine) # 㳰
  when "LTS2GZ","LTS2GW","LTS2VZ","LTS2VW","LTSRVZ","LTSRVW"
    vname = "im"
    invars.push(vname)
    declare += "    int #{vname}1;\n"
    r2c = "    #{vname}1 = NUM2INT(#{vname});\n" + r2c
  end


  case(outvars.length)
  when(0)
    outlist = "0"
  when(1)
    outlist = outvars[0]
  else
    outlist = "rb_ary_new3(#{outvars.length}, #{outvars.join(", ")})"
  end

  file << "\n"
  file << "static VALUE\n"
#  file << "#{subroutine.downcase}(#{inlist})\n"
#  file << "    VALUE #{inlist};\n"
  file << "#{subroutine.downcase}(argc,argv,self)\n"
  file << "    int argc; VALUE *argv, self;\n"
  file << "{\n"
  file << "    VALUE #{invars.join(",")};\n" if invars.length > 0
  file << declare
  file << "\n"
  file << declare_c2r if declare_c2r
  file << "\n"
  for i in 0..invars.length-1
    file << "    #{invars[i]} = argv[#{i}];\n"
  end
  file << "\n"
  file << r2c
  file << "\n"
  file << "    #{subroutine.downcase}_(#{args.chop});\n"
  file << "\n"
  file << c2r
  file << "\n"
  file << "    return #{outlist};\n"
  file << "}\n"

#  p subroutine.downcase+"( "+inlist+" )"

#  definelist += "    rb_define_module_function(m#{modulename},\"#{subroutine.downcase}\",#{subroutine.downcase},#{inlist.split(",").length-1});\n"
  definelist += "    rb_define_module_function(m#{modulename},\"#{subroutine.downcase}\",#{subroutine.downcase},-1);\n"

} # end of routinelist.each


file << "\n"
file << "\n"
file << "Init_#{libfilename}()\n"
file << "{\n"
file << "    VALUE m#{modulename} = rb_define_module(\"#{modulename}\");\n"
file << definelist
file << "}\n"

file.close

end # if(cwrapper) # Cåѡ񤭽Ф
#-----------------------------------------------------------------------------#
if(rubywrapper) # Ruby̥åѡǤ񤭽Ф(ޤʲϤΤ)

file = File.open(rbfilename,"w")

classname = ""

routinelist.sort.each{ |subroutine|
  if varlist[subroutine] == nil
    raise "varlist for #{subroutine} not exist"
  end

  if classname != subroutine[0..1].upcase
    file << "end\n" unless classname == ""
    classname = subroutine[0..1].upcase
    file << "\n"
    file << "class #{subroutine[0..1].upcase}PACK\n"
  end

  argvars = Array.new
  invars = Array.new
  outvars = Array.new
  walloc = ""

  vars1[subroutine].each{|v|

    vname = v["name"]
    if v["inout"]=="in" || v["inout"]=="inout"

      case(vname)
      when "im","jm","km","lm","mm","q","r","it","t","iti","ti","itj","tj","itk","tk","m","n","nm"
        invars.push("@"+vname)
      else
        argvars.push(vname)
        invars.push(vname)
      end
    end
    if v["inout"]=="work"
      arylength = v["type"].downcase.sub(/\(i/,"").sub(/\(d/,"").sub(/\)$/,"").gsub(/([^a-z])([a-z])/,'\1'+"@"+'\2')
      walloc += "    _alloc_#{vname}(#{arylength})\n"
      invars.push("@"+vname)
    end
    if v["inout"]=="out" || v["inout"]=="inout"
      outvars.push(vname)

    end

  } # end of vars1[subroutine].each

  if subroutine[2..5] == "INIT"
    file << "  def initialize(#{argvars.join(", ")})\n"
    file << "    ISPACK::#{subroutine.downcase}(#{invars.join(",")})\n"
    file << "  end\n"
  else
    file << "  def #{subroutine[2..5].downcase}(#{argvars.join(", ")})\n"
    file << walloc
    file << "    ISPACK::#{subroutine.downcase}(#{invars.join(",")})\n"
    file << "  end\n"
  end

} # end of routinelist.each

file << "end\n"

file.close

end # if(rubywrapper) # Ruby̥åѡ񤭽Ф
#-----------------------------------------------------------------------------#
